import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

DEBUG = True
TEMPLATE_DEBUG = True

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
    }
}

TIME_ZONE = "Europe/Berlin"
LANGUAGE_CODE = "en-us"

SITE_ID = 1

USE_I18N = True
USE_L10N = True
MEDIA_ROOT = os.path.join(BASE_DIR, "media")
STATIC_ROOT = os.path.join(BASE_DIR, "static")
MEDIA_URL = "/media/"
STATIC_URL = "/static/"
SECRET_KEY = '0pfuvtvasasd8623123foobar76723"b)lna4*f_-xxkszs4##!+wpo'
ROOT_URLCONF = "example.urls"
TEMPLATE_DIRS = (os.path.join(BASE_DIR, "templates"),)

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "django_video_encoder",
    "example",
]

DJANGO_VIDEO_ENCODER_THUMBNAIL_INTERVAL = 20
DJANGO_VIDEO_ENCODER_FORMATS = [
    {"label": "H.264 (HD)", "codec": "h264"},
    {"label": "H.264", "codec": "h264", "width": 720, "height": 404},
    {"label": "VP9 (HD)", "codec": "VP9"},
    {"label": "VP9", "codec": "VP9", "width": 720, "height": 404},
]



from django.contrib.contenttypes.fields import GenericRelation
from django.db import models

from django_video_encoder.models import Format


class Video(models.Model):
    video_file = models.FileField()

    formats = GenericRelation(Format)

    class Meta:
        app_label = "example"



# Generated by Django 2.2.16 on 2020-09-30 12:56

import django.db.models.deletion
from django.db import migrations, models

import django_video_encoder.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Thumbnail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                ("time", models.PositiveIntegerField(verbose_name="Time (s)")),
                ("width", models.PositiveIntegerField(null=True, verbose_name="Width")),
                (
                    "height",
                    models.PositiveIntegerField(null=True, verbose_name="Height"),
                ),
                (
                    "image",
                    models.ImageField(
                        blank=True,
                        max_length=512,
                        null=True,
                        upload_to=django_video_encoder.models.thumbnail_upload_to,
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Format",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                ("field_name", models.CharField(max_length=255)),
                (
                    "format",
                    models.CharField(choices=[("H.264", "H.264")], max_length=255),
                ),
                (
                    "file",
                    models.FileField(
                        max_length=2048,
                        upload_to=django_video_encoder.models.format_upload_to,
                    ),
                ),
                ("width", models.PositiveIntegerField(null=True, verbose_name="Width")),
                (
                    "height",
                    models.PositiveIntegerField(null=True, verbose_name="Height"),
                ),
                (
                    "duration",
                    models.PositiveIntegerField(
                        null=True, verbose_name="Duration (ms)"
                    ),
                ),
                (
                    "extra_info",
                    models.TextField(
                        blank=True, verbose_name="Encoder information (JSON)"
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
        ),
    ]



import logging

from django.conf import settings
from django.core import signing
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from .tasks import get_video_task

logger = logging.getLogger(__name__)


@csrf_exempt
@require_POST
def notification(request):
    if getattr(settings, "ZENCODER_NOTIFICATION_SECRET", None) and (
        request.META.get("HTTP_X_ZENCODER_NOTIFICATION_SECRET")
        != settings.ZENCODER_NOTIFICATION_SECRET
    ):
        logger.warning(
            "Invalid Zencoder notification secret", extra={"request": request}
        )
        return HttpResponse("Invalid notification secret", status=400)  # BAD REQUEST

    try:
        data = signing.loads(request.META["QUERY_STRING"])
    except signing.BadSignature:
        logger.warning(
            "Invalid payload for Zencoder notification", extra={"request": request}
        )
        return HttpResponse("Invalid payload", status=400)  # BAD REQUEST

    get_video_task.delay(
        data["ct"], data["obj"], data["fld"], request.body.decode("utf-8")
    )
    return HttpResponse(status=204)  # NO CONTENT



from django.dispatch import Signal

sent_to_zencoder = Signal(providing_args=["instance", "result"])
sending_failed = Signal(providing_args=["instance", "error"])
encoding_failed = Signal(providing_args=["instance", "result"])
received_format = Signal(providing_args=["instance", "format", "result"])



from django.contrib.contenttypes import admin

from .models import Format


class FormatInline(admin.GenericTabularInline):
    model = Format
    fields = ("format", "file", "width", "height", "duration")
    readonly_fields = fields
    extra = 0
    max_num = 0



import os.path

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.utils.text import slugify

from .api import encode

VIDEO_ENCODER_MODELS = {}


def format_upload_to(instance, filename):
    dirname, original_filename = os.path.split(
        getattr(instance.video, instance.field_name).name
    )
    return "formats/%s/%s/%s%s" % (
        instance.format,
        dirname or ".",
        slugify(os.path.splitext(original_filename)[0]),
        os.path.splitext(filename)[1].lower(),
    )


class Format(models.Model):
    object_id = models.PositiveIntegerField()
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)

    video = GenericForeignKey()
    field_name = models.CharField(max_length=255)

    format = models.CharField(
        max_length=255,
        choices=[
            (f["label"], f["label"]) for f in settings.DJANGO_VIDEO_ENCODER_FORMATS
        ],
    )
    file = models.FileField(upload_to=format_upload_to, max_length=2048)
    width = models.PositiveIntegerField("Width", null=True)
    height = models.PositiveIntegerField("Height", null=True)
    duration = models.PositiveIntegerField("Duration (ms)", null=True)

    extra_info = models.TextField("Encoder information (JSON)", blank=True)


def thumbnail_upload_to(instance, filename):
    return "footage/thumbnails/%s/%s.jpg" % (
        instance.video.pk,
        instance.time,
    )


class Thumbnail(models.Model):
    object_id = models.PositiveIntegerField()
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)

    time = models.PositiveIntegerField("Time (s)")
    width = models.PositiveIntegerField("Width", null=True)
    height = models.PositiveIntegerField("Height", null=True)
    image = models.ImageField(
        upload_to=thumbnail_upload_to, max_length=512, blank=True, null=True
    )

    video = GenericForeignKey()


def detect_file_changes(sender, instance, **kwargs):
    field_name = VIDEO_ENCODER_MODELS.get(
        "%s.%s" % (sender._meta.app_label, sender._meta.model_name)
    )
    if field_name:
        field = getattr(instance, field_name)
        if field and hasattr(field, "file") and isinstance(field.file, UploadedFile):
            if hasattr(instance, "_video_encoder_updates"):
                instance._video_encoder_updates.append(field_name)
            else:
                instance._video_encoder_updates = [field_name]


def trigger_encoding(sender, instance, **kwargs):
    for field in getattr(instance, "_video_encoder_updates", ()):
        encode(instance, field)


if getattr(settings, "VIDEO_ENCODER_MODELS", None):
    for name in settings.VIDEO_ENCODER_MODELS:
        app_model, field = name.rsplit(".", 1)
        VIDEO_ENCODER_MODELS[app_model.lower()] = field
    models.signals.pre_save.connect(detect_file_changes)
    models.signals.post_save.connect(trigger_encoding)



import cgi
import datetime
import json
import logging
import os
from os.path import basename
from urllib.error import URLError
from urllib.request import Request, urlopen, urlretrieve

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.core import signing
from django.core.exceptions import ObjectDoesNotExist
from django.core.files import File
from django.urls import reverse

from . import signals
from .errors import ZencoderError

logger = logging.getLogger(__name__)


def open_url(url, data=None):
    if data:
        headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }
        request = Request(url, data=json.dumps(data).encode("utf-8"), headers=headers)
    else:
        request = Request(url)

    try:
        response = urlopen(request)
    except URLError as e:
        raise ZencoderError(e.reason)

    if response.getcode() // 100 != 2:
        try:
            raise ZencoderError(", ".join(json.loads(response.text)["errors"]))
        except ValueError:
            raise ZencoderError(response.reason or "HTTP error: %d" % response.status)

    return response


def send_request(data):
    data["api_key"] = settings.ZENCODER_API_KEY
    response = open_url("https://app.zencoder.com/api/v2/jobs", data)
    return json.loads(response.read().decode("utf-8"))


def encode(obj, field_name, file_url=None):
    def absolute_url(url):
        """
        Helper to turn a domain-relative URL into an absolute one
        with protocol and domain
        """
        domain = Site.objects.get_current().domain
        protocol = (
            "https" if getattr(settings, "ZENCODER_NOTIFICATION_SSL", False) else "http"
        )
        return url if "://" in url else "%s://%s%s" % (protocol, domain, url)

    if not file_url:
        file_url = getattr(obj, field_name).url

    content_type = ContentType.objects.get_for_model(type(obj))

    color_metadata = "preserve"
    if getattr(settings, "ZENCODER_DISCARD_COLOR_METADATA", "preserve"):
        color_metadata = "discard"

    outputs = []
    for fmt in settings.DJANGO_VIDEO_ENCODER_FORMATS:
        data = {
            "obj": obj.pk,
            "ct": content_type.pk,
            "fld": field_name,
        }
        notification_url = "%s?%s" % (
            absolute_url(reverse("zencoder_notification")),
            signing.dumps(data),
        )

        outputs.append(
            {
                "label": fmt["label"],
                "video_codec": fmt["codec"],
                "width": fmt.get("width"),
                "height": fmt.get("height"),
                "notifications": [notification_url],
                "color_metadata": color_metadata,
            }
        )

    data = {
        "input": absolute_url(file_url),
        "region": getattr(settings, "ZENCODER_REGION", "europe"),
        "output": outputs,
        "test": getattr(settings, "ZENCODER_INTEGRATION_MODE", False),
    }

    # get thumbnails for first output only
    data["output"][0]["thumbnails"] = {
        "interval": settings.DJANGO_VIDEO_ENCODER_THUMBNAIL_INTERVAL,
        "start_at_first_frame": 1,
        "format": "jpg",
    }

    try:
        result = send_request(data)
    except ZencoderError as e:
        result = None
        logger.warning(
            "Error when sending encoding request to zencoder for %s/%s/%s: %s",
            content_type,
            obj.pk,
            field_name,
            e,
        )
        signals.sending_failed.send(sender=type(obj), instance=obj, error=e)
    else:
        logger.info(
            "Sent encoding request for %s/%s/%s, job id: %s",
            content_type,
            obj.pk,
            field_name,
            result["id"],
        )
        signals.sent_to_zencoder.send(sender=type(obj), instance=obj, result=result)
    return result


def get_video(content_type_id, object_id, field_name, data):
    content_type = ContentType.objects.get(id=content_type_id)
    logger.info("Getting video file for %s/%s/%s", content_type, object_id, field_name)

    output = json.loads(data)["output"]

    try:
        obj = content_type.get_object_for_this_type(pk=object_id)
    except ObjectDoesNotExist:
        logger.warning(
            "The model %s/%s has been removed after being sent to Zencoder",
            content_type,
            object_id,
            field_name,
        )
    else:
        if output["state"] == "finished":

            from .models import Format, Thumbnail

            # get preview pictures
            if output.get("thumbnails"):
                for i, thumbnail in enumerate(output["thumbnails"][0]["images"]):
                    filename, header = urlretrieve(thumbnail["url"])
                    thmb, __ = Thumbnail.objects.get_or_create(
                        content_type=content_type,
                        object_id=object_id,
                        time=i * settings.DJANGO_VIDEO_ENCODER_THUMBNAIL_INTERVAL,
                    )
                    thmb.image.save(basename(filename), File(open(filename, "rb")))
                    os.unlink(filename)

            fmt, __ = Format.objects.get_or_create(
                content_type=content_type,
                object_id=object_id,
                field_name=field_name,
                format=output["label"],
            )

            response = open_url(output["url"])
            try:
                # parse content-disposition header
                filename = cgi.parse_header(response.info()["Content-Disposition"])[1][
                    "filename"
                ]
            except (KeyError, TypeError):
                filename = "format_%s.%s" % (
                    datetime.datetime.now().strftime("%Y%m%d_%H%M%S"),
                    response.info()["Content-Type"].rsplit("/", 1)[1],
                )

            # remove trailing parameters
            filename = filename.split("?", 1)[0]

            f = File(response)
            f.size = response.info()["Content-Length"]

            fmt.width = output["width"]
            fmt.height = output["height"]
            fmt.duration = output["duration_in_ms"]
            fmt.extra_info = data
            fmt.file.save(basename(filename), f)
            logger.info(u"File %s saved as %s", filename, fmt.file.name)
            signals.received_format.send(
                sender=type(obj), instance=obj, format=fmt, result=data
            )

        elif output["state"] == "failed":
            logger.warning(
                "Zencoder error for %s/%s/%s: %s",
                content_type,
                object_id,
                field_name,
                output["error_message"],
            )
            signals.encoding_failed.send(sender=type(obj), instance=obj, result=data)

        else:
            logger.error(
                "Unknown zencoder status for %s/%s/%s: %s",
                content_type,
                object_id,
                field_name,
                data,
            )



from unittest import mock

from django.core.files import File
from pytest import mark

from django_video_encoder.api import encode
from example.models import Video


@mark.django_db
def test_encode_called(mocker):
    """Assert encode will try to send data via the api"""
    file_mock = mock.MagicMock(spec=File)
    file_mock.name = "video.vid"
    mocked_send_request = mocker.patch("django_video_encoder.api.send_request")
    video_obj = Video(video_file=file_mock)
    encode(video_obj, "video_file")
    assert mocked_send_request.called



from setuptools import find_packages, setup

setup(
    name="django_video_encoder",
    description="Simple integration of a Django model with a FileField to encoding backends",
    long_description=open("README.rst").read(),
    author="Jonas und der Wolf GmbH",
    author_email="opensource@jonasundderwolf.de",
    url="http://github.com/jonasundderwolf/django-video-encoder",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



# Generated by Django 2.2.16 on 2020-10-22 16:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_video_encoder", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="format",
            name="format",
            field=models.CharField(
                choices=[
                    ("H.264 (HD)", "H.264 (HD)"),
                    ("H.264", "H.264"),
                    ("VP9 (HD)", "VP9 (HD)"),
                    ("VP9", "VP9"),
                ],
                max_length=255,
            ),
        ),
    ]
